@@ -279,6 +279,7 @@ gksu_context_get_debug (GksuContext *context)
  * Set up if we are running inside a sshd forwarding
  * session. That's because the X authorization semantics
  * differ a bit.
+ * -> THIS SETTING IS CURRENTLY USED FOR NOTHING
  */
 void
 gksu_context_set_ssh_fwd (GksuContext *context, gboolean value)
@@ -292,6 +293,7 @@ gksu_context_set_ssh_fwd (GksuContext *context, gboolean value)
  *
  * Finds out if the library is configured as being run inside a
  * ssh X11 forwarding tunnel.
+ * -> THIS SETTING IS CURRENTLY USED FOR NOTHING
  *
  * Returns: TRUE if it is, FALSE otherwise.
  */
@@ -418,18 +420,14 @@ prepare_xauth (GksuContext *context)
   gchar **tmpv = NULL;
   gchar *display = NULL;
 
-  if (context->ssh_fwd)
-    {
-      display = g_strdup (getenv ("DISPLAY"));
-      tmpv = g_strsplit (display, ":", 3);
-      g_free (display);
-
-      context->display = g_strdup_printf (":%s", tmpv[1]);
-
-      g_strfreev (tmpv);
-    }
-  else
-    context->display = g_strdup (getenv ("DISPLAY"));
+  /* avoid problems with "network" DISPLAY's */
+  display = g_strdup (getenv ("DISPLAY"));
+  tmpv = g_strsplit (display, ":", 3);
+  g_free (display);
+  
+  context->display = g_strdup_printf (":%s", tmpv[1]);
+  
+  g_strfreev (tmpv);
 
   /* get the authorization token */
   tmp = g_strdup_printf ("/usr/X11R6/bin/xauth list %s | "
@@ -443,7 +441,7 @@ prepare_xauth (GksuContext *context)
 	       strerror(errno));
       return 1;
     }
-  fread (xauth, sizeof(char), 255, xauth_output);
+  fread (xauth, sizeof(char), 256, xauth_output);
   xauth[strlen(xauth) - 1] = '\0';
   pclose (xauth_output);
   g_free (tmp);
@@ -745,6 +743,8 @@ gksu_context_run (GksuContext *context, GError **error)
 
 	      if (!first_line_read)
 		{
+		  read (fdpty, buf, 255);
+
 		  if (!strncmp (buf, "su: Authentication failure", 26))
 		    auth_failed = TRUE;
 		  else if (!strncmp (buf, "gksu: waiting", 13))
@@ -759,9 +759,10 @@ gksu_context_run (GksuContext *context, GError **error)
 		      write (fdpty, line, strlen(line));
 		      g_free (line);
 
-		      first_line_read = TRUE;
 		      bzero (buf, 255);
 		    }
+
+		  first_line_read = TRUE;
 		}
 	      else if (!strncmp (buf, "gksu-run:", 8))
 		{
@@ -842,7 +843,6 @@ gksu_context_sudo_run (GksuContext *context, GError **error)
   pid_t pid;
   int status;
   size_t r;
-  char buf[16];
   FILE *infile, *outfile;
   int parent_pipe[2];	/* For talking to the parent */
   int child_pipe[2];	/* For talking to the child */
@@ -877,14 +877,48 @@ gksu_context_sudo_run (GksuContext *context, GError **error)
 
   cmd = g_new (gchar *, argcount + 1);
   
-  cmd[0] = g_strdup("/usr/bin/sudo"); /* sudo binary */
-  cmd[1] = g_strdup("-H"); /* Make sudo set $HOME */
-  cmd[2] = g_strdup("-S"); /* Make sudo read from stdin */
-  cmd[3] = g_strdup("-p"); /* Make sudo use next arg as prompt */
-  cmd[4] = g_strdup("GNOME_SUDO_PASS ");	/* prompt */
-  cmd[5] = g_strdup("-u"); /* Make sudo use the selected user */
-  cmd[6] = g_strdup(context->user);    /* user */
-  cmd[7] = g_strdup("--"); /* Make sudo stop processing options */  
+  argcount = 0;
+
+  /* sudo binary */
+  cmd[argcount] = g_strdup("/usr/bin/sudo");
+  argcount++;
+
+  if (!context->keep_env)
+    {
+      /* Make sudo set $HOME */
+      cmd[argcount] = g_strdup("-H");
+      argcount++;
+    }
+
+  /* Make sudo read from stdin */
+  cmd[argcount] = g_strdup("-S");
+  argcount++;
+
+  /* Make sudo use next arg as prompt */
+  cmd[argcount] = g_strdup("-p");
+  argcount++;
+
+  /* prompt */
+  cmd[argcount] = g_strdup("GNOME_SUDO_PASS");
+  argcount++;
+
+  /* Make sudo use the selected user */
+  cmd[argcount] = g_strdup("-u");
+  argcount++;
+
+  /* user */
+  cmd[argcount] = g_strdup(context->user);
+  argcount++;
+
+  /* sudo does not understand this if we do not use -H
+     weird.
+  */
+  if (!context->keep_env)
+    {
+      /* Make sudo stop processing options */
+      cmd[argcount] = g_strdup("--");
+      argcount++;
+    }
 
   for (i = j = 0; ; i++)
     {
@@ -961,16 +995,11 @@ gksu_context_sudo_run (GksuContext *context, GError **error)
     } 
   else 
     {
+      gboolean auth_failed = FALSE;
+
       // Parent
       close(parent_pipe[1]);
       
-      /*
-	we are expecting to receive a GNOME_SUDO_PASS
-	if we don't there are two possibilities: an error
-	or a password is not needed
-      */
-      r = read(parent_pipe[0], buf, 16);
-
       infile = fdopen(parent_pipe[0], "r");
       if (!infile)
 	{
@@ -991,8 +1020,21 @@ gksu_context_sudo_run (GksuContext *context, GError **error)
 	  return -1;
 	}
 
-      if (strncmp(buf, "GNOME_SUDO_PASS ", 16) == 0) 
+      /*
+	we are expecting to receive a GNOME_SUDO_PASS
+	if we don't there are two possibilities: an error
+	or a password is not needed
+      */
+      r = read(parent_pipe[0], buffer, 16);
+
+      if (context->debug)
+	fprintf (stderr, "buffer: -%s-\n", buffer);
+
+      if (strncmp(buffer, "GNOME_SUDO_PASS", 16) == 0) 
 	{
+	  if (context->debug)
+	    fprintf (stderr, "Yeah, we're in...\n");
+
 	  if (context->password == NULL)
 	    {
 	      g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_NOPASSWORD,
@@ -1005,24 +1047,38 @@ gksu_context_sudo_run (GksuContext *context, GError **error)
 	  fprintf (outfile, "%s", context->password);
 	  fclose (outfile);
 	} 
-      else
-	fprintf (stderr, "%s", buf);
+      else 
+	{
+	  if (context->debug)
+	    fprintf (stderr, "Oops... what's up?\n");
+
+	  buffer[16] = 0;
+	  fprintf (stderr, "%s", buffer);
+	}
+
+      /* ignore the newline that comes right after sudo receives
+	 the password
+      */
+      fgets (buffer, 255, infile);
+      fgets (buffer, 255, infile);
+      if (!strcmp (buffer, "Sorry, try again.\n"))
+	auth_failed = TRUE;
 
       while (!waitpid (pid, &status, WNOHANG))
 	{
-	  bzero(buf, 256);
-	  if(!fread (buf, sizeof(gchar), 255, infile))
+	  bzero(buffer, 256);
+	  if(!fgets (buffer, 255, infile))
 	    break;
-	  fprintf (stderr, "%s", buf);
+	  fprintf (stderr, "%s", buffer);
 	}
 
       /* make sure we did read everything */
       while (1)
 	{
-	  bzero(buf, 256);
-	  if(!fread (buf, sizeof(gchar), 255, infile))
+	  bzero(buffer, 256);
+	  if(!fread (buffer, sizeof(gchar), 255, infile))
 	    break;
-	  fprintf (stderr, "%s", buf);
+	  fprintf (stderr, "%s", buffer);
 	  fflush (stderr);
 	}
       
@@ -1030,11 +1086,16 @@ gksu_context_sudo_run (GksuContext *context, GError **error)
 	{
 	  if (WEXITSTATUS(status))
 	    {
-	      g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_CHILDFAILED,
-			   _("Child terminated with %d status"),
-			   WEXITSTATUS(status));
+	      if (auth_failed)
+		g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_WRONGPASS,
+			     _("Wrong password."));
+	      else
+		g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_CHILDFAILED,
+			     _("Child terminated with %d status"),
+			     WEXITSTATUS(status));
+
 	      sudo_reset_xauth (context, xauth, xauth_env);
-	      return -1;
+	      return TRUE;
 	    }
 	}
       
@@ -1042,6 +1103,6 @@ gksu_context_sudo_run (GksuContext *context, GError **error)
     }
   
   sudo_reset_xauth (context, xauth, xauth_env);
-  return 0;
+  return FALSE;
 }
 
